import logging
from random import choice
import string

logger = logging.getLogger('dna.util.stats')
logger.addHandler(logging.NullHandler())

def random_string(l=5):
    return ''.join([choice(string.ascii_letters + string.digits) for n in range(l)])

class dnastats:
    """ collect stats regarding the dnastorage system simulation """
    def __init__(self, msg="", fd=None):
        self.fd = fd
        self.msg = msg
        self.all_stats = {}
        self.formats = {}

    def set_fd(self, fd):
        self.fd = fd

    # Use inc to track and dump a counter. example:
    #    stats.inc("block::errors")
    # If called 5 times, this will produce a log entry like this:
    #    block::errors=5
    #
    def inc(self, name, i=1, dflt=0):
        self.all_stats[name] = self.all_stats.get(name,dflt) + i

    # Use append to record a list of data
    def append(self, name, s, dflt=[]):
        self.all_stats[name] = self.all_stats.get(name,dflt) + [s]

    def unique(self, name, val):
        while name in self.all_stats:
            name += "::"+random_string(5)
        self.all_stats[name] = val            
        
    def __getitem__(self, name):
        return self.all_stats[name]

    def __setitem__(self, name, value):
        self.all_stats[name] = value

    def format(self, name, f):
        self.formats[name] = f

    def persist(self):
        if not (self.fd is None):
            if len(self.msg) > 0:
                self.fd.write(self.msg+"\n")
                logger.info(self.msg)

            items = self.all_stats.items()
            items.sort()
                
            for k,v in items:
                fmt = "{}="+"{}\n".format(self.formats.get(k,"{}"))
                self.fd.write(fmt.format(k,v))
                logger.info(fmt.format(k,v))
        else:
            if len(self.msg) > 0:
                logger.info(self.msg)
            items = self.all_stats.items()
            items.sort()
            for k,v in items:
                fmt = "{}="+"{}".format(self.formats.get(k,"{}"))
                logger.info(fmt.format(k,v))
                                
    def __del__(self):
        return

global stats
stats = dnastats(msg="global stats collection for dnastorage project")

if __name__ == "__main__":
    import sys
    from dnastorage.util.stats import dnastats,stats
    logger.setLevel(logging.DEBUG)
    _formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    _ch = logging.FileHandler("dna.debug.log",mode='w')
    _ch.setFormatter(_formatter)
    logger.addHandler(_ch)
    stats.inc("unique_strands")
    stats.inc("unique_strands")
    stats.inc("unique_strands")
    



import editdistance as ed
from io import BytesIO
import math
import struct

from dnastorage.codec.base_conversion import convertIntToBytes,convertBytesToInt
from dnastorage.codec.builder import customize_RS_CFC8
from dnastorage.util.packetizedfile import ReadPacketizedFilestream, WritePacketizedFilestream
import dnastorage.system.formats as formats

### Designed to fit on a single strand for most use cases
### 
### Every header strand begins with special sequence that can't be used at the beginning of indices: ATCGATGC
###
### 1. 'ATCGATGC' [1]
### 2. short index - usually 0, 0-255, at most 16 strands [1]
### 3. major version (0-255) [1]
### 4. minor version (0-255) [1]
### 5. num bytes for size [1]
### 6. size [x]
### 7  num bytes for original filename [2]
### 8. null terminated string
### 9. encoding style        [2]
### 10. length of remaining record (2 bytes)      [2]
### 11. remaining record byte encoded  [?]
### Pad to final width using arbitrary sequence


system_version = { 'major': 0, 'minor':1 }
magic_header = 'ATCGATGC' #'CCATCCAT'

def encode_primer_diff(o,n):
    hdr = []
    baseVal = { 'A': 0, 'C': 1, 'G':2, 'T':3 }                            
    assert len(o)==len(n)
    for i,(oo,nn) in enumerate(zip(o,n)):
        if oo != nn:                                
            hdr +=  [ (baseVal[nn] << 6) | (i&0x3F) ]
    if len(hdr) == 0:
        return [0]
    hdr = [len(hdr)] + hdr
    return hdr

def decode_primer_diff(data,oprimer):
    sz = data[0]
    if sz==0:
        return oprimer,1    
    baseVal = [ 'A', 'C', 'G', 'T' ]
    nprimer = [ _ for _ in oprimer ]
    for i in range(sz):
        val = data[1+i]
        base = baseVal[(val&0xC0)>>6]
        pos = val&0x3F
        nprimer[pos] = base
    return "".join(nprimer),sz+1

def encode_size_and_value(val):
    data = []
    if val==0:
        data += [1, 0]
    else:
        data += convertIntToBytes(int(math.ceil(math.log(val+1,2)/8.0)),1)
        data += convertIntToBytes(int(val),int(math.ceil(math.log(val+1,2)/8.0)))
    return data

def decode_size_and_value(data,pos):
    #print "decode_size_and_value: ",pos, len(data)
    size_bytes = data[pos]
    val = convertBytesToInt(data[pos+1:pos+1+size_bytes])
    return val,size_bytes+1


def encode_file_header_comments(filename,format_id,size,other_data,primer5,primer3):
    comment = "% dnastorage version {}.{}\n".format(system_version['major'],system_version['minor'])
    comment += "% {} \n".format(size)
    if len(filename) > 0:
        comment += "% {} \n".format(filename)
    else:
        comment += "% No filename recorded.\n"

    comment += "% 5-{} 3-{}\n".format(primer5,primer3)
    comment += "% Id-{} Description-{} \n".format(format_id,formats.file_system_format_description(format_id))
    comment += "% {} bytes of additional data \n".format(len(other_data))
    return comment

def encode_file_header(filename,format_id,size,other_data,primer5,primer3,fsmd_abbrev='FSMD'):
    data =  [ system_version['major'], system_version['minor'] ]
    data += convertIntToBytes(int(math.ceil(math.log(size,2)/8.0)),1)
    data += convertIntToBytes(size,int(math.ceil(math.log(size,2)/8.0)))
    data += convertIntToBytes(len(filename)+1,2)
    data += [ ord(_) for _ in filename ] + [0]        
    data += convertIntToBytes(format_id,2)
    data += convertIntToBytes(len(other_data),2)
    data += other_data
    #data += [0]*(80-len(data))
    #data = bytearray([x for x in data])
    data = bytearray(data)

    #print "size of file header: ",len(data)
    
    pf = ReadPacketizedFilestream(BytesIO(data))
    enc_func = formats.file_system_encoder_by_abbrev(fsmd_abbrev)
    enc = enc_func(pf,primer5+magic_header,primer3)
    strands = []
    for e in enc:
        if type(e) is list:
            for s in e:                
                strands.append(s)
        else:
            strands.append(e)
            
    return strands

def pick_nonheader_strands(strands,primer5):
    others = []
    picks = []
    for s in strands:
        if s.startswith(primer5):            
            if s.startswith(primer5+magic_header):
                pass
            else:
                others.append(s)
        else:
            others.append(s)
    return others

def pick_header_strands(strands,primer5):
    picks = []
    others = []
    for s in strands:
        if s.find(primer5+magic_header)!=-1:
            picks.append(s)
        elif s.find(primer5)!=-1:
            plen= s.find(primer5)+len(primer5)
            possible_hdr = s[plen:plen+len(magic_header)]
            #if ed.eval(possible_hdr,magic_header) < 2:
            #    print "guessed at header!"
                #ss = s[:]
                #ss[plen:plen+len(magic_header)] = magic_header
            #    picks.append(s)
            #else:
            others.append(s)

    return picks,others


def decode_file_header(strands,primer5,primer3,fsmd_abbrev='FSMD'):
    picks,_ = pick_header_strands(strands,primer5)

    b = BytesIO()

    fid = formats.file_system_formatid_by_abbrev(fsmd_abbrev)
    packetsize = formats.file_system_format_packetsize(fid)

    pf = WritePacketizedFilestream(b,packetsize,packetsize)
    
    dec_func = formats.file_system_decoder_by_abbrev(fsmd_abbrev)
    dec = dec_func(pf,primer5+magic_header,primer3)

    #d = {}
    for s in picks:
        #tmp = dec.decode_from_phys_to_strand(s)
        #d[tmp[0]] = d.get(tmp[0],[])
        #d[tmp[0]].append(tmp)
        #print len(tmp),tmp
        dec.decode(s)

    #print d

    dec.write()
    assert dec.complete    

    #print (b.getvalue())
    try:
        data = [ ord(x) for x in b.getvalue() ]
    except Exception as e:
        data = [ x for x in b.getvalue() ]
        
    #print data
    
    assert data[0] == system_version['major']
    assert data[1] == system_version['minor']

    header = {}
    header['version'] = [ data[0], data[1] ]

    size_bytes = data[2]
    pos = 3
    header['size'] = convertBytesToInt(data[pos:pos+size_bytes])
    pos += size_bytes

    size_filename = convertBytesToInt(data[pos:pos+2])
    pos+=2
    header['filename'] = "".join([chr(x) for x in data[pos:pos+size_filename]])
    pos += size_filename

    header['formatid'] = convertBytesToInt(data[pos:pos+2])
    pos += 2
    
    size_other_data = convertBytesToInt(data[pos:pos+2])
    pos += 2
    header['other_data'] = [ x for x in data[pos:pos+size_other_data] ]

    #print "size_other_data={}".format(size_other_data)
    #print "len(other_data)={}".format(len(header['other_data']))
    
    
    return header

if __name__ == "__main__":
    strands = encode_file_header("",0xA,2,[1,2,3,4],"A"*19+"G","T"*19+"G")
    for s in strands:
        print ("{}: strand={}".format(len(s), s))
    
    print (decode_file_header(strands,"A"*19+"G","T"*19+"G"))



#!/usr/bin/python
import os
import sys

import logging
logger = logging.getLogger("dna.storage.util.packetizedfile")
logger.addHandler(logging.NullHandler())

class WritePacketizedFilestream:
    """
    Write a file by receiving (index,value) packets of data of a specific size at a given index. Value
    is of packetSize length.  index specifies where in the file to write the data.  Packets can be
received in any order. But, the file cannot be written until the file is complete, in other words,
    it has received all size/packetSize packets.  The indices are assumed to be densely numbered from 0
    to size/packetSize by 1.  Pythonically: range(0,size/packetSize+(size-size/packetSize*packetSize),1).
    """
    def __init__(self,fd,size,packetSize,minKey=0,zeroFillMissing=False):
        self.__fd = fd
        self.size = size
        self.__set_packetSize(packetSize)
        self.__data = {}
        self.minKey = int(minKey)                      # inclusive
        self.zeroFillMissing = zeroFillMissing
        
    def has_key(self,key):
        return key in self.__data
    
    def __setitem__(self,key,value):        
        if (key >= self.minKey) and (key < self.maxKey):
            self.__data[key] = value
        else:
            print ("not in range:",key,value,self.minKey,self.maxKey)
    def __getitem__(self,key):
        assert key >= self.minKey and key < self.maxKey
        return self.__data[key]

    @property
    def maxKey(self):
        return int(self.minKey+self.numberOfPackets) # exclusive
    
    @property
    def numberOfPackets(self):
        if (self.size % self.packetSize) > 0:
            return int(self.size / self.packetSize + 1)
        else:
            return int(self.size / self.packetSize)

    # packet property
    def __set_packetSize(self,val):
        self.__packetSize = val
    def __get_packetSize(self):
        return self.__packetSize
    packetSize = property(__get_packetSize,__set_packetSize)

    @property
    def lastPacketSize(self):
        # this function should never return 0, not the same as modulo
        # should return a number >= 1 and <= self.packetSize
        return self.size - (self.numberOfPackets-1)*self.packetSize

    @property
    def complete(self):
        if len(self.__data.keys()) < self.numberOfPackets:
            #print "too few keys {}".format(self.numberOfPackets)
            return False
        for i in range(self.minKey,self.maxKey):
            if not (i in self.__data.keys()):
                #print "missing key {}".format(i)
                return False
        return True

    def getMissingKeys(self):
        keys = self.__data.keys()
        missing = []
        for i in range(self.minKey,self.maxKey):
            if not i in keys:
                missing.append(i)
        return missing


    ## Warning: requires buffering the whole file!
    def write(self):
        #emptyString = '\x00'*self.packetSize
        emptyString = bytearray(self.packetSize)
        for i in range(self.minKey,self.maxKey):
            #print (self.__data[i])
            if i in self.__data:
                if i == self.maxKey-1:
                    self.__fd.write(self.__data[i][0:self.lastPacketSize])
                else:
                    self.__fd.write(self.__data[i])
            else:
                if self.zeroFillMissing:
                    self.__fd.write(emptyString)
                
        # for key,value in items:
        #     if i < key:
        #         while i < key:
        #             self.__fd.write(emptyString)
        #             i+=1
        #     if i == self.numberOfPackets-1:
        #         self.__fd.write(value[0:self.lastPacketSize])
        #     else:
        #         self.__fd.write(value)
        #     i+=1
        self.__fd.flush()
            
    def close(self):
        if self.__fd != sys.stdout and self.__fd != sys.stderr:
            self.__fd.close()

    def dummy_write(self):
        items = self.__data.items()
        items.sort(cmp=lambda x,y: cmp(x[0],y[0]))
        i = 0
        emptyString = '\x00'*self.packetSize
        output_data=""
        for key,value in items:
            if value is not type(str):
                value2="".join(value)
            else:
                value2 = value
            if i < key:
                while i < key and i < self.numberOfPackets-1:
                    output_data+=emptyString
                    i+=1
            if i == self.numberOfPackets-1:
                output_data+=value2[0:self.lastPacketSize]
                i+=1
                break
            else:
                output_data+=value2
                i+=1
            if i==self.numberOfPackets:
                break
        while(i<self.numberOfPackets):
            if i == self.numberOfPackets-1:
                output_data+=emptyString[0:self.lastPacketSize]
            else:
                output_data+=emptyString
            i+=1
        #print i,self.numberOfPackets,len(output_data),self.size
        assert i == self.numberOfPackets and len(output_data)==self.size 
        return output_data



class WritePacketizedFile(WritePacketizedFilestream):
    def __init__(self,filename,size,packetSize):
        WritePacketizedFilestream.__init__(self,open(filename,"wb"),size,packetSize)
        self.__filename = filename



class ReadPacketizedFilestream:
    """
    Read a file by breaking it up into packetSize pieces. If the last piece is smaller
    than packetSize, it's padded with zeros. Clients are guaranteed all packets are uniform
    size.
    
    This is an iterable object. Packets can be read using the iterator or by requesting
    a specific index between 0 and size/packetSize.
    
    This class expects a file descriptor. There is a derived class that accepts a filename.
    """
    def __init__(self,fd):
        self.__fd = fd
        self.__set_packetSize(120)
        self.__read_size = 0
        self._RS=False

    def read(self):
        b = bytes(self.__fd.read(self.packetSize))
        self.__read_size += len(b)
        #only pad out if it is not RS, RS encodes by blocks so padding may lead to many useless strands
        if b and len(b) != self.packetSize and not self._RS:
            b = b.ljust(self.packetSize,bytes(1))

        return b

    # packet property
    def __set_packetSize(self,val):
        self.__packetSize = val
    def __get_packetSize(self):
        return self.__packetSize
    packetSize = property(__get_packetSize,__set_packetSize)

    # iterator
    def __iter__(self):
        # reset to begnning of file
        self.__fd.seek(0)
        return self

    def __next__(self):
        b = self.read()
        if b:
            return b
        else:
            raise StopIteration()

    next = __next__  # preserve python 2.7 functionality
        
    def __getitem__(self,key):
        self.__fd.seek(key*self.packetSize)
        return self.read()

    @property
    def size(self):
        return os.fstat(self.__fd.fileno()).st_size
    @property
    def bytes_read(self):
        return self.__read_size

    @property
    def numberOfPackets(self):
        if (self.size % self.packetSize)==0:
            return self.size / self.packetSize
        else:
            return self.size / self.packetSize + 1


class ReadPacketizedFile(ReadPacketizedFilestream):
    def __init__(self,filename):
        ReadPacketizedFilestream.__init__(self,open(filename,"rb"))
        self.__filename = filename
    @property
    def filename(self):
        return self.__filename


if __name__ == "__main__":
    import os
    import sys
    from random import randint
    packetFile = ReadPacketizedFile(sys.argv[1])

    out = WritePacketizedFile("output.d",packetFile.size,120)

    assert out.complete==False

    i = 0
    for p in packetFile:
        out[i] = p
        i += 1

    assert out.complete==True
    out.write()




from dnastorage.util.stats import stats

class FaultTolerancePolicy(object):
    """ Describes policy for handling faults """
    def __init__(self):
        pass
    def allow(self, e):
        """ Override this method to give exception specific behavior """
        assert False and "Not implemented"
        return True
    def __str__(self):
        return "No policy specified"

class AllowAll(FaultTolerancePolicy):
    """ Describes policy for handling faults """
    def __init__(self):
        super(AllowAll,self).__init__()
        pass
    def allow(self, e):
        return True
    def __str__(self):
        return "Tolerate all faults as well as possible, attempt to handle"

class NoTolerance(FaultTolerancePolicy):
    def __init__(self):
        super(NoTolerance,self).__init__()
        pass
    def allow(self, e):
        return False
        

# All exceptions in dnastorage should inherit from this class
class DNAStorageError(Exception):
    """ Exception base class for error handling in this module. All """
    """ exceptions should derive from here. """
    def __init__(self,msg=None):
        if msg is None:
            msg = "A failure of some kind occurred in the dnastorage module."
        super(DNAStorageError,self).__init__(msg)
        stats.inc("Error")

class DNACodingError(DNAStorageError):
    """ An error occured while encoding or decoding a file """
    def __init__(self,msg=None):
        if msg is None:
            msg = "An error occurred while encoding/decoding a file."
        super(DNACodingError,self).__init__(msg)
        stats.inc("CodingError")


class DNAReedSolomonOuterCodeError(DNACodingError):
    """ An error occured while encoding or decoding a file """
    def __init__(self,msg=None):
        if msg is None:
            msg = "An error occurred while encoding/decoding a file."
        super(DNAReedSolomonOuterCodeError,self).__init__(msg)
        stats.inc("RSOuterCodeError")

class DNABlockBadIndex(DNACodingError):
    """ An error occured while decoding a block """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Block had the wrong inter-block index."
        super(DNABlockBadIndex,self).__init__(msg)
        stats.inc("BlockBadIndex")

class DNABlockMissingIndex(DNACodingError):
    """ An error occured while decoding a block """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Block is missing an index."
        super(DNABlockMissingIndex,self).__init__(msg)
        stats.inc("BlockMissingIndex")

class DNABlockTooLargeError(DNACodingError):
    """ An error occured while decoding a block """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Block was too large."
        super(DNABlockTooLargeError,self).__init__(msg)
        stats.inc("BlockTooLargeError")


class DNABlockPayloadWrongSize(DNACodingError):
    """ An error occured while decoding a block """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Strand in the block was the wrong size."
        super(DNABlockPayloadWrongSize,self).__init__(msg)
        stats.inc("BlockPayloadWrongSize")


class DNABadCodeword(DNACodingError):
    """ An error occured while decoding a codeword symbol """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Strand in the block was the wrong size."
        super(DNABadCodeword,self).__init__(msg)
        stats.inc("BadCodeword")

class DNAStrandPayloadWrongSize(DNACodingError):
    """ An error occured while decoding a strand """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Strand was the wrong size."
        super(DNAStrandPayloadWrongSize,self).__init__(msg)
        stats.inc("StrandPayloadWrongSize")


class DNAStrandMissingSequence(DNACodingError):
    """ Strand is missing a special sequence (like a cut) """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Strand was missing a special sequence."
        super(DNAStrandMissingSequence,self).__init__(msg)
        stats.inc("StrandMissingSequence")

class DNAMissingPrimer(DNACodingError):
    """ Strand is missing a primer """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Strand was missing a primer."
        super(DNAMissingPrimer,self).__init__(msg)
        stats.inc("MissingPrimer")


class DNAStrandPoorlyFormed(DNACodingError):
    """ Strand is poorly formed """
    def __init__(self,msg=None):
        if msg is None:
            msg = "Strand is poorly formed."
        super(DNAStrandPoorlyFormed,self).__init__(msg)
        stats.inc("StrandPoorlyFormed")

class DNAFileHeaderHasError(DNACodingError):
    """ Header is poorly formed """
    def __init__(self,msg=None):
        if msg is None:
            msg = "DNAFile header has an error."
        super(DNAFileHeaderHasError,self).__init__(msg)
        stats.inc("HeaderHasError")
        



from io import BytesIO
import sys
import logging

from dnastorage.util.packetizedfile import ReadPacketizedFilestream, WritePacketizedFilestream
import dnastorage.system.formats as formats
import dnastorage.system.header as header

logger = logging.getLogger('dna.storage.system.dnafile')
logger.addHandler(logging.NullHandler())

class DNAFile:
    def __init__(self):
        return

    @classmethod
    def open(self, filename, op, primer5, primer3, format_name="", write_incomplete_file=False, fsmd_abbrev='FSMD',flanking_primer5="",flanking_primer3="",use_flanking_primer_for_decoding=False):
        '''        
        1. filename is the input file of strands
        2. op is the file operation: 'r' or 'w'
        3. format_name is optional for reading and required for writing. 
        4. primer5 is the coding primer, and primer3 is the non-coding primer.
        '''     
        # check if we are reading or writing
        if op=="r":
            
            with open(filename,"r") as fd:
                logger.debug("open {} for reading.".format(filename))
                strands = get_strands(fd)

                if use_flanking_primer_for_decoding==True:
                    h = header.decode_file_header(strands,flanking_primer5+primer5,\
                                       flanking_primer3+primer3,fsmd_abbrev=fsmd_abbrev)
                else:
                    h = header.decode_file_header(strands,primer5,primer3,fsmd_abbrev=fsmd_abbrev)
            
                logger.debug("decoded header: {}".format(h)) 
                assert h['version'][0] <= header.system_version['major']
                assert h['version'][1] <= header.system_version['minor']
            
                return ReadDNAFile(input=filename, primer5=primer5, primer3=primer3,\
                                   fsmd_abbrev=fsmd_abbrev,\
                                   flanking_primer5=flanking_primer5,\
                                   flanking_primer3=flanking_primer3,\
                                   use_flanking_primer_for_decoding=use_flanking_primer_for_decoding)

        elif op=="w":
            return WriteDNAFile(output=filename,primer5=primer5,primer3=primer3, \
                                format_name=format_name,fsmd_abbrev=fsmd_abbrev,\
                                flanking_primer5=flanking_primer5,\
                                flanking_primer3=flanking_primer3)
        else:
            return None

    def flush(self):
        return

    def close(self):
        return

    def readable(self):
        assert False
    def writable(self):
        assert False

def get_strands(in_fd):
    strands = []
    while True:
        s = in_fd.readline()        
        if len(s) == 0:
            break
        s = s.strip()
        if s.startswith('%'):
            continue
        strands.append(s)
    return strands

class ReadDNAFile(DNAFile):
    # ReadDNAFile reads a set of strands from a file.  It finds the header,
    # determines compatibility and encoding type, and then decodes the file.
    #
    def __init__(self,**kwargs):     
        DNAFile.__init__(self)

        if 'input' in kwargs:
            self.input_filename = kwargs['input']
            self.in_fd = open(self.input_filename,"r")
            need_to_close = True
        elif 'in_fd' in kwargs:
            self.in_fd = kwargs['in_fd']
            self.input_filename = ""
            need_to_close = False
            
        if not ('fsmd_abbrev' in kwargs):
            self.fsmd_abbrev = 'FSMD'
        else:
            self.fsmd_abbrev = kwargs['fsmd_abbrev']
            
        assert 'primer5' in kwargs and 'primer3' in kwargs
        self.primer5 = kwargs['primer5']
        self.primer3 = kwargs['primer3']

        if 'flanking_primer5' in kwargs:
            self.flanking_primer5 = kwargs['flanking_primer5']
        else:
            self.flanking_primer5 = ''

        if 'flanking_primer3' in kwargs:
            self.flanking_primer3 = kwargs['flanking_primer3']
        else:
            self.flanking_primer3 = ''

        if 'use_flanking_primer_for_decoding' in kwargs:
            self.use_flanking_primers = kwargs['use_flanking_primer_for_decoding']
        else:
            self.use_flanking_primers = False
            
        # get all the strands ( this will be bad for large files )
        strands = get_strands(self.in_fd)

        h = header.decode_file_header(strands,self.primer5,self.primer3,fsmd_abbrev=self.fsmd_abbrev)

        self.strands = header.pick_nonheader_strands(strands,self.primer5)
        
        assert h['version'][0] <= header.system_version['major']
        assert h['version'][1] <= header.system_version['minor']

        self.formatid = h['formatid']
        self.header = h 
        self.size = h['size']

        # set up mem_buffer 
        self.mem_buffer = BytesIO()
        
        if self.formatid == 0x1000:
            # let sub-classes handle initialization
            return

        dec_func = formats.file_system_decoder(self.formatid)
            
            
        self.mem_buffer = BytesIO()
        self.pf = WritePacketizedFilestream(self.mem_buffer,self.size,formats.file_system_format_packetsize(self.formatid))

        if self.use_flanking_primers:
            self.dec = dec_func(self.pf,self.flanking_primer5+self.primer5,\
                                self.flanking_primer3+self.primer3)
        else:
            self.dec = dec_func(self.pf,kwargs['primer5'],kwargs['primer3'])

        for s in self.strands:
            if s.startswith(self.primer5):
                self.dec.decode(s)

        self.dec.write()
        assert self.dec.complete

        self.mem_buffer.seek(0,0) # set read point at beginning of buffer
        
        if need_to_close:
            self.in_fd.close()
        return

    def read(self, n=1):        
        return self.mem_buffer.read(n)
    def readline(self, n=-1):
        return self.mem_buffer.readline(n)

    def readable(self):
        return True
    def writable(self):
        return False
        
class WriteDNAFile(DNAFile):
    # WriteDNAFile writes a set of strands.
    def __init__(self,**kwargs):     
        DNAFile.__init__(self)
        if 'formatid' in kwargs:            
            enc_func = formats.file_system_encoder(kwargs['formatid'])
            self.formatid = kwargs['formatid']
        elif 'format_name' in kwargs:
            enc_func = formats.file_system_encoder_by_abbrev(kwargs['format_name'])
            self.formatid = formats.file_system_formatid_by_abbrev(kwargs['format_name'])

        if not ('fsmd_abbrev' in kwargs):
            self.fsmd_abbrev = 'FSMD'
        else:
            self.fsmd_abbrev = kwargs['fsmd_abbrev']

        self.mem_buffer = BytesIO()
        self.pf = ReadPacketizedFilestream(self.mem_buffer)

        assert 'primer5' in kwargs and 'primer3' in kwargs
        self.primer5 = kwargs['primer5']
        self.primer3 = kwargs['primer3']

        if 'flanking_primer3' in kwargs:
            self.flanking_primer3 = kwargs['flanking_primer3']
        else:
            self.flanking_primer3 = ''

        if 'flanking_primer5' in kwargs:
            self.flanking_primer5 = kwargs['flanking_primer5']
        else:
            self.flanking_primer5 = ''

        self.enc = enc_func(self.pf,self.flanking_primer5+self.primer5,\
                            self.flanking_primer3+self.primer3)

            
        if 'output' in kwargs:
            self.output_filename = kwargs['output']
            self.out_fd = open(self.output_filename,"w")
        elif 'out_fd' in kwargs:
            self.out_fd = kwargs['out_fd']
            self.output_filename = ""
            
        self.size = 0
        self.strands = []
        return

    def _encode_buffer(self):
        for block in self.enc:
            if type(block) == list:
                for s in block:
                    self.strands.append(s)
            else:
                self.strands.append(block)
        #print "_encode_buffer: index={}".format(self.enc.index)

    def read(self, size):
        assert False and "Do not read at the same time as writing"
        
    def writable(self):
        return True
    def readable(self):
        return False
    
    def write(self, buff):
        self.size += len(buff)
        tell = self.mem_buffer.tell()
        self.mem_buffer.seek(0,2)

        #print ( len(buff), bytes([x for x in buff]), buff )

        #buff = bytes([x for x in buff])        
        #buff = bytes([ord(x) for x in buff])
        try:
            # convert string
            self.mem_buffer.write(buff)
            self.mem_buffer.seek(tell,0)
        except Exception as e:
            buff = bytearray([x for x in buff])
            self.mem_buffer.write(buff)
            self.mem_buffer.seek(tell,0)
            
        return
    
    def flush(self):
        self._encode_buffer()
        return

    def close(self):
        self.flush()
        hdr = header.encode_file_header(self.output_filename,self.formatid,self.size,\
                                 "",self.flanking_primer5+self.primer5,\
                                 self.flanking_primer3+self.primer3,\
                                 fsmd_abbrev=self.fsmd_abbrev)

        for i,h in enumerate(hdr):
            self.strands.insert(i,h)

        comment = header.encode_file_header_comments(self.output_filename,self.formatid,self.size,"",\
                                              self.primer5,self.primer3)
        self.out_fd.write(comment)
        for s in self.strands:
            self.out_fd.write("{}\n".format(s))

        if self.out_fd != sys.stdout and self.out_fd != sys.stderr:
            self.out_fd.close()
        return


    



import random
import editdistance as ed

import dnastorage.exceptions as err
from dnastorage.codec.base_codec import BaseCodec

class CombineCodewords(BaseCodec):
    def __init__(self,CodecObj=None,Policy=None):
        BaseCodec.__init__(self,CodecObj=CodecObj,Policy=Policy)

    def _encode(self, codeword_list):
        return "".join(codeword_list)

    def _decode(self, s):
        assert ("not used for decoding")


class NormalizeStrandLength(BaseCodec):
    def __init__(self,length,CodecObj=None,Policy=None):
        BaseCodec.__init__(self,CodecObj=CodecObj,Policy=Policy)
        self.length = length
        
    def _encode(self, phys_s):
        if len(phys_s) > self.length:
            e = err.DNAStrandPayloadWrongSize("NormalizeStrandLength: Strand is too long ({})".format(len(phys_s)))
            if self._Policy.allow(e):
                pass
            else:
                raise e
        elif len(phys_s) < self.length:
            add = self.length - len(phys_s)
            phys_s = phys_s + "".join([ random.choice('AGCT') for _ in range(add) ])
            
        return phys_s

    def _decode(self, s):
        assert ("not used for decoding")

        
class InsertMidSequence(BaseCodec):
    def __init__(self,seq,CodecObj=None,Policy=None):
        BaseCodec.__init__(self,CodecObj=CodecObj,Policy=Policy)
        self._seq = seq

    def _encode(self,strand):
        if strand.find(self._seq) != -1:
            e = err.DNAStrandPoorlyFormed("Found sequence already present while prepending {}"\
                                        .format(self._seq))
            if self._Policy.allow(e):
                pass
            else:
                raise e

        middle = int(len(strand)/2)
        return strand[0:middle]+self._seq+strand[middle:]

    def _decode(self,strand):
        index = strand.find(self._seq)
        if index != -1:
            return strand[0:index]+strand[index+len(self._seq):]
        else:
            er = err.DNAStrandMissingSequence("{} should have had {} inside it.".format(strand,self._seq))
            # there could be errors in the cut preventing us from seeing it
            # with an exact match, so now we look for an inexact match
            if self._Policy.allow(er):
                middle = int(len(strand)/2)
                slen = len(self._seq)
                res = []
                for m in range(middle-slen,middle):
                    sli = strand[m:m+slen]
                    res.append( ed.eval(sli,self._seq) )
                mn = min(res)
                if mn < slen/3:
                    place = middle - slen + res.index(mn)
                    return strand[0:place]+strand[place+slen:]
                else:
                    # just leave the strand along, and hopefully codewords can
                    # still be extracted properly
                    return strand
            else:
                raise er


class PrependSequence(BaseCodec):
    def __init__(self,seq,CodecObj=None,Policy=None,isPrimer=False):
        BaseCodec.__init__(self,CodecObj=CodecObj,Policy=Policy)
        self._seq = seq[:]
        self.is_primer = isPrimer

    def _encode(self,strand):
        if strand.find(self._seq) != -1:
            er = err.DNAStrandPoorlyFormed("Found sequence already present while prepending {}"\
                                        .format(self._seq))
            if self._Policy.allow(er):
                pass
            else:
                raise er
        return self._seq + strand

    def _decode(self,strand):
        index = strand.find(self._seq)
        if index != -1: # expected at beginning
            return strand[index+len(self._seq):]
        else:
            er = err.DNAStrandMissingSequence("{} should have had {} at beginning.".format(strand,self._seq))
            # there could be errors in the cut preventing us from seeing it
            # with an exact match, so now we look for an inexact match
            if self._Policy.allow(er):
                slen = len(self._seq)
                res = []
                # FIXME: how far in should we look?
                for m in range(0,50):
                    sli = strand[m:m+slen]
                    res.append( ed.eval(sli,self._seq) )
                mn = min(res)
                
                idx = res.index(mn)
                #print (res,mn)
                if mn < 5:
                    return strand[idx+slen:]
                else:
                    if self.is_primer:
                        raise err.DNAMissingPrimer("Missing primer {}".format(self._seq))
                    # just leave the strand along, and hopefully codewords can
                    # still be extracted properly
                    return strand
            else:
                raise er


class AppendSequence(BaseCodec):
    def __init__(self,seq,CodecObj=None,Policy=None,isPrimer=False):
        BaseCodec.__init__(self,CodecObj=CodecObj,Policy=Policy)
        self._seq = seq
        self.is_primer = isPrimer

    def _encode(self,strand):
        if strand.find(self._seq) != -1:
            er = err.DNAStrandPoorlyFormed("Found sequence already present while appending {}"\
                                        .format(self._seq))
            if self._Policy.allow(er):
                pass
            else:
                raise er
        
        return strand + self._seq

    def _decode(self,strand):
        index = strand.find(self._seq)
        slen = len(self._seq)
        if index != -1: # expected at end
            return strand[:index]
        else:
            er = err.DNAStrandMissingSequence("{} should have had {} at end.".format(strand,self._seq))
            # there could be errors in the cut preventing us from seeing it
            # with an exact match, so now we look for an inexact match
            if self._Policy.allow(er):
                slen = len(self._seq)
                res = []
                for m in range(len(strand)-2*slen,len(strand)):
                    sli = strand[m:m+slen]
                    res.append( ed.eval(sli,self._seq) )
                mn = min(res)
                idx = res.index(mn)
                if mn < 5:
                    return strand[:idx+len(strand)-2*slen]
                else:
                    if self.is_primer:
                        raise err.DNAMissingPrimer("Missing primer".format(self._seq))
                    # just leave the strand along, and hopefully codewords can
                    # still be extracted properly
                    return strand
            else:
                raise er


if __name__ == "__main__":
    import random
    
    cut = InsertMidSequence('AGATATAGGG')
    pre = PrependSequence('TAAAGGAAAAAG',CodecObj=cut)
    app = AppendSequence( 'CAAAATATAAAA',CodecObj=pre)
    app = app
    
    match = 0
    for _ in range(10000):
        strand = [ random.choice('AGCT') for _ in range(100) ]
        strand = "".join(strand)

        original = strand

        strand = app.encode(strand)

        copy = [ random.choice('AGCT') for _ in range(10) ] + [ _ for _ in strand ] + [ random.choice('AGCT') for _ in range(10) ]
        copy[ random.randint(0,9) ] = random.choice('AGCT')
        copy[ random.randint(140,150) ] = random.choice('AGCT')
        copy[ random.randint(75,78) ] = random.choice('AGCT')
        copy = "".join(copy)

        
        new =  app.decode(copy)
        if new == original:
            match += 1
        else:
            print ("--",len(original)-len(new))
            print (original)
            print (new)
            
    print (match / 10000.0 * 100)



#This file contains code for generating random numbers based on specified parameters, uses the inverse transform method with cutpoints to speed up generation
from scipy import special
from scipy import stats
import random
import math
import numpy




#Instantiate this class in order to generate random variables from the negative binomial distribution with gamma function parameterization (rather than classical parameterization)
class neg_bin:
    def __init__(self,mean,var):
        self._cumulative_array=[]
        self._prob_array=[]
        #variance should be greater than the mean for neg binomial, otherwise use Poisson distribution
        assert var>mean
        self._theta=mean**2/(var-mean)
        self._mean=mean
        self._var=var
        last_cumlative=0
        index=0
        #build the cummulative probability array and probability array 
        while last_cumlative < 99.999E-2:
            prob=self.pmf(index)
            self._prob_array.append(prob)
            if not self._cumulative_array:
                self._cumulative_array.append(prob)
            else:
                self._cumulative_array.append(self._cumulative_array[-1]+prob)
            if self._cumulative_array[-1]>=1:
                break
            index=index+1
            last_cumlative=self._cumulative_array[-1]
        if self._cumulative_array[-1]<1:
            self._cumulative_array[-1]=1

        #need to define cutoff points used for quickly getting random values
        self._m=len(self._cumulative_array)
        k=-1
        m_q=0
        self._cutoff_array=[]
        for j in range(0,self._m):
            while m_q<=j:
                k=k+1
                m_q=self._m*self._cumulative_array[k]
            self._cutoff_array.append(k)
            
        assert len(self._cumulative_array)==self._m
        
        
    #calculate a probability given a value
    def pmf(self,X):
        theta=self._theta
        mean=self._mean
        #gammaln is the natural log of the gamma function at the specified value, used to make sure the value from the gamma function does not explode
        log_prob=special.gammaln(X+theta)-special.gammaln(X+1)-special.gammaln(theta)+theta*(math.log(theta)-math.log(mean+theta))+X*(math.log(mean)-math.log(theta+mean))
        prob = math.exp(log_prob)
        return prob

    #function for generating negative binomial random value from a uniform random variable
    def gen(self):
        #Use this uniform random variable to 
        U=random.uniform(0,1)
        cumulative_index=self._cutoff_array[int(math.floor(U*self._m))]
        while U>self._cumulative_array[cumulative_index]:
            cumulative_index=cumulative_index+1
        return cumulative_index

    def get_mean(self):
        return self._mean

    def get_var(self):
        return self._var

    

def bins_array(data,bin_size):
    num_bins=int(math.ceil((max(data)-min(data))/bin_size))
    lower_bin=min(data)
    upper_bin=lower_bin+bin_size
    bin_array={}
    for bin_number in range(0,num_bins):
        frequency=0
        for count in data:
            if count>=lower_bin and count<upper_bin:
                frequency=frequency+1
        bin_array[lower_bin]=frequency
        lower_bin=upper_bin
        upper_bin=upper_bin+bin_size
    return bin_array,num_bins





def calculate_expected(num_samples,start,bin_size,prob_array,num_bins):
    expected={}
    index=start
    for i in range(0,num_bins):
        expected[int(index)]=num_samples*sum(prob_array[index:index+bin_size])
        index=index+bin_size
    return expected





    
if __name__ == "__main__":
    import matplotlib.pyplot as plt

    _rand=neg_bin(200,3000)
    rand_reads=[]
    for i in range (0,1000000):
        rand_reads.append(int(_rand.gen()))

    #keep bin size integer, since we have a discrete random variable


    bin_size=1


    bins_,num_bins=bins_array(rand_reads,bin_size)
    start_point=min(rand_reads)
    probability_array=_rand._prob_array
    num_samples=len(rand_reads)
    expected=calculate_expected(num_samples,start_point,bin_size,probability_array,num_bins)
    assert len(expected)==len(bins_)
    sorted_bin_results=[bins_[x] for x in sorted(bins_)]
    sorted_expected_results=[expected[x] for x in sorted(expected)]
    chi_dist=0
    for index, value in enumerate(sorted_bin_results):
        #print "{}   {}".format(sorted_bin_results[index],sorted_expected_results[index])
        chi_dist=chi_dist+((value-sorted_expected_results[index])**2/(sorted_expected_results[index]))
        #print"chi {}".format(chi_dist)
    print (chi_dist)
    print (len(sorted_bin_results)-1-2)
    print (1 - stats.chi2.cdf(chi_dist, len(sorted_bin_results)-1-2))

    
    plt.hist(rand_reads,num_bins)
    plt.show()



from dnastorage.codec.base_codec import BaseCodec

class EncodePacketizedFile(object):
    """
    Partition the file into key,value packets for encoding. This is an iterable
    object for convenient encoding. If a CodecObj is provided, the iterator
    returns the packets after they pass through the CodecObj.

    For simple encoders and decoders that operate on one strand at a time, you can simply
    override the _encode function to alter how encoding is done before the CodecObj is invoked.

    For customized keys, you need to override the class and alter how keys are
    generated.

    """

    def __init__(self,packetizedFile,CodecObj=None,minIndex=0):
        self._index = minIndex
        #print "self._index={}".format(self._index)
        self.minIndex = minIndex
        self._packetizedFile = packetizedFile
        self._iterating = False
        if CodecObj == None:
            self._Codec = BaseCodec()
        else:
            self._Codec = CodecObj

    @property
    def index(self):
        #print "@property _index={}".format(self._index)
        return self._index
            
    @property
    def bytes_encoded(self):
        return self._packetizedFile.bytes_read

    # Derived classes should override only this method, unless other
    # functionality also needs to be altered
    # returns key, value
    def _encode(self):
        #print "return self._index={}".format(self._index)
        return self._index,self._packetizedFile.next()

    # Ideally, this would not be overriden, but if a subclass wants to
    # fully control all steps of encoding, it must be overriden
    def encode(self):
        return self._Codec.encode(self._encode())

    # Sub-classes that alter key names, may need to override this
    # Right now, only using one file descriptor, so this will
    # cause iteration to break. Either use this interface or iteration,
    # not both!
    def __getitem__(self,key):
        if self._iterating:
            assert False and "Warning: should not use __getitem__ while iterating!"
        return self._packetizedFile[key]

    # iterator
    def __iter__(self):
        self._index = self.minIndex
        self._iterating = True
        self._packetizedFile.__iter__()
        # reset to begnning of file
        return self

    def __next__(self):
        packet = self.encode()
        if packet:
            self._index += 1
            return packet
        else:
            self._iterating = False
            raise StopIteration()

    next = __next__ # support for python 2.7
        
class DecodePacketizedFile(object):
    def __init__(self,packetizedFile,CodecObj=None,minIndex=0):
        self._packetizedFile = packetizedFile
        self._data = {}
        self.minIndex = minIndex
        if CodecObj == None:
            self._Codec = BaseCodec()
        else:
            self._Codec = CodecObj

    def has_key(self,key):
        return self._data.has_key(key)

    #def __setitem__(self,key,value):
    #    self._data[key] = value
    #def __getitem__(self,key):
    #    return self._data[key]

    def writeToFile(self,key,value):
        #print (key, len(value), value, type(value))
        # make sure value is a string
        # try:
        #     if (type(value) is list) and (type(value[0]) is int or type(value[0]) is long):
        #         value = "".join([chr(x) for x in value])
            
        #     if (type(value) is list) and (type(value[0]) is str):
        #         value = "".join([x for x in value])
        #     # assert type(value) is str
        #     self._packetizedFile[key] = value
        #except Exception as e:
        if (type(value) is list) and (type(value[0]) is int or type(value[0]) is long):
            value = bytearray(value)
        elif (type(value) is list) and (type(value[0]) is str):
            value = bytearray([ord(x) for x in value])
        else:
            value = bytearray(value)
        self._packetizedFile[key] = value
                
    # Ideally, derived classes will only override this implementation
    def _decode(self,key,value):
        self.writeToFile(key,value)

    def decode(self,strand,bypass=False,input_key=None,input_value=None):
        #if input_key < self.minIndex or input_key >= self._packetizedFile.maxKey:
        #    return
        if bypass is False:
            key,val = self._Codec.decode(strand)
        else:
            key=input_key
            val=input_value
        #print key,val
        self._decode(key,val)


    @property
    def complete(self):
        return self._packetizedFile.complete

    def write(self):
        self._packetizedFile.write()

    def dummy_write(self):
        return self._packetizedFile.dummy_write()


if __name__ == "__main__":
    import os
    import sys
    from random import randint
    from dnastorage.util.packetizedfile import pf
    packetFile = pf.ReadPacketizedFile(sys.argv[1])
    out = pf.WritePacketizedFile("output.d",packetFile.size,120)
    assert out.complete==False
    i = 0
    for p in packetFile:
        out[i] = p
        i += 1
    assert out.complete==True
    out.write()



from random import randint

import dnastorage.exceptions as err
from dnastorage.codec.base_codec import BaseCodec
from dnastorage.codec.reedsolomon.rs import ReedSolomon,get_reed_solomon,ReedSolomonError
from dnastorage.util.stats import stats

import logging
logger = logging.getLogger('dna.storage.codec.strand')
logger.addHandler(logging.NullHandler())

class RandomizeCodec(BaseCodec):
    '''Add numRandBytes random values to the end of the strand. This is a naive way 
       of ensuring distinct strands during voting.
    '''
    def __init__(self,numRandBytes,CodecObj=None,Policy=None):
        super(RandomizeCodec,self).__init__(CodecObj=CodecObj,Policy=Policy)
        self._numRandBytes = numRandBytes

    def _encode(self, array):
        for i in range(self._numRandBytes):
            array.append( randint(0,255) )

    def _decode(self, array):
        return array[:self._numRandBytes]

class ReedSolomonInnerCodec(BaseCodec):
    """
    ReedSolomonInnerCodec takes a sequence of bytes as input to the _encode function and
    produces a Reed-Solomon encoded message as a byte array. 

    This is an 'Inner' Codec because it only can correct errors within a strand.
    """
    def __init__(self,numberECCBytes,c_exp=8,CodecObj=None,Policy=None):
        """
        numberECCBytes is the amount of error correction symbols to add. 
        c_exp is the characteristic.
        CodecObj is a nested codec, called after encoding or before decoding. See BaseCodec 
        for more info.
        Policy defines the fault handling support. By default, errors trigger an exception.        
        """
        super(ReedSolomonInnerCodec,self).__init__(CodecObj=CodecObj,Policy=Policy)

        self.rs = get_reed_solomon(c_exp=c_exp)
        self._numberECCBytes = numberECCBytes

    def _encode(self,array):
        """ Accepts list/array of bytes. Return the Reed-Solomon encoded byte array.
        """
        try:
            assert len(array) <= self.rs.field_charac
        except:
            #print "Failed RS check"
            return array
        # usually not needed, but makes the code a little more tolerant for use with
        # a variety of codecs that may pass in strings, bytearrays, or lists:
        message = [x for x in array]

        try:
            # encoded the message using the RS library
            mesecc = self.rs.rs_encode_msg(message, self._numberECCBytes)
        except ReedSolomonError as e:
            raise err.DNACodingError("Error while encoding Reed-Solomon Inner Codec.")
        except ZeroDivisionError as e:
            pass
        
        #print "RSInner:",len(mesecc)
        return mesecc

    def _decode(self,array):
        """
        This function expects a list of unsigned integers in the GF. For now, erasures are
        denoted with -1.
        """
        message = [x for x in array] 
        # find the -1s in the list
        erasures = [ i for i in range(len(message)) if message[i]==-1 ]
        # correct the message
        try:
            corrected_message, corrected_ecc = self.rs.rs_correct_msg(message,self._numberECCBytes, erase_pos=erasures)
            value = corrected_message
            #print "corrected message"
            stats.inc("RSInnerCodec::decode::succeeded")
        except ReedSolomonError as e:
            stats.inc("RSInnerCodec::decode::failed")
            #print "Inner: Couldn't correct message: {}".format(message)
            stats.inc("RSInnerCodec.ReedSolomonError")
            if self._Policy.allow(e):
                # leave erasures, may be helpful for outer decoder
                value = message[0:(self._numberECCBytes)]
                pass # nothing to do
            else:
                print (str(e))
                raise err.DNACodingError("RSInnerCodec failed to correct message.")
            # just proceed without further error correction
            pass
        except ZeroDivisionError as e:
            stats.inc("RSInnerCodec.ZeroDivision")
            if self._Policy.allow(e):
                # leave erasures, may be helpful for outer decoder
                value = message[0:(self._numberECCBytes)]
                pass # nothing to do
            else:
                print (str(e))
                raise err.DNACodingError("RSInnerCodec failed to correct message.")

        return value



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dnastorage", 
    version="0.9.1",
    author="James M. Tuck",
    author_email="james.m.tuck@gmail.com",
    description="dnastorage provides basic support for encoding and decoding files into and out of DNA with several convenient interfaces.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dna-storage/dnastorage",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    python_requires='>=3.6',
)



from random import randint
from math import ceil,log

import dnastorage.exceptions as err
from dnastorage.codec.codecfile import EncodePacketizedFile, DecodePacketizedFile
from dnastorage.codec.phys import CombineCodewords
from dnastorage.codec.block import doMajorityVote, partitionStrandsIntoBlocks
from dnastorage.codec.block import reportBlockStatus

from dnastorage.util.stats import stats

class LayeredEncoder(EncodePacketizedFile):

    def __init__(self,packetizedFile,minIndex=0,\
                 strandSizeInBytes=10,blockSizeInBytes=200*10,\
                 blockCodec=None,\
                 blockToStrandCodec=None,\
                 strandCodec=None,\
                 strandToCodewordCodec=None,\
                 codewordToPhysCodec=CombineCodewords(),\
                 physCodec=None,\
                 Policy=None):
        
        EncodePacketizedFile.__init__(self,packetizedFile,minIndex=minIndex)
        # set packetSize (can't hurt to do it again here)
        self._packetizedFile.packetSize = blockSizeInBytes

        # make sure sizes are consistent with each other
        # blocks must be whole multiple of strand size
        assert blockSizeInBytes % strandSizeInBytes == 0

        self.blockCodec = blockCodec
        self.blockToStrandCodec = blockToStrandCodec
        self.strandCodec = strandCodec
        self.strandToCodewordCodec = strandToCodewordCodec
        self.codewordToPhysCodec = codewordToPhysCodec
        self.physCodec = physCodec
                
        self.strandSizeInBytes = strandSizeInBytes
        self.blockSizeInBytes = blockSizeInBytes
        return

    def _layered_encode(self, block):
        # perform block encoding
        enc_block = self.blockCodec.encode(block)

        # convert block into strands
        strands = self.blockToStrandCodec.encode(enc_block)


        # protected logical strands
        final_strands = []
        tmp_strands = []
        for s in strands:
            ecc_s = self.strandCodec.encode(s)
            tmp_strands.append(ecc_s)
            cw_s = self.strandToCodewordCodec.encode(ecc_s)
            # phys codecs expect a DNA sequnce as a string
            phys_s = self.codewordToPhysCodec.encode(cw_s)            
            final = self.physCodec.encode(phys_s)
            final_strands.append(final)

        return final_strands

    
    def dummy_encode(self):
        """ Dummy encode let's us run the sequence of codecs and produce a single
            strand. We can look at that strand to determine its length or possibly
            other properties. Helpful when designing codecs.
        """
        block = [ randint(0,255) for _ in range(self.blockSizeInBytes) ]
        strands = self._layered_encode( (randint(0,255**2),block) )
        return strands
                
    def encode(self):        
        block = self._encode()
        try:
            block = (block[0],[ ord(_) for _ in block[1] ])
        except Exception as e:
            # This is pretty horrible, kids don't try this at home!
            block = (block[0],[ _ for _ in block[1] ])
        return self._layered_encode(block) # get entire block
        

class LayeredDecoder(DecodePacketizedFile):
    
    def __init__(self,packetizedFile,minIndex=0,\
                 strandSizeInBytes=10,blockSizeInBytes=200*10,\
                 blockIndexSize=2,
                 physCodec=None,\
                 physToStrandCodec=None,\
                 strandCodec=None,\
                 strandToBlockCodec=None,\
                 blockCodec=None,\
                 Policy=None,
                 intraBlockIndexSize=1):\
                         
        DecodePacketizedFile.__init__(self,packetizedFile,minIndex=minIndex)
        # set packetSize (can't hurt to do it again here)
        self._packetizedFile.packetSize = blockSizeInBytes

        # make sure sizes are consistent with each other
        # blocks must be whole multiple of strand size
        assert blockSizeInBytes % strandSizeInBytes == 0

        self.physCodec = physCodec
        self.physToStrandCodec = physToStrandCodec
        self.strandCodec = strandCodec
        self.strandToBlockCodec = strandToBlockCodec        
        self.blockCodec = blockCodec
                
        self.strandSizeInBytes = strandSizeInBytes
        self.blockSizeInBytes = blockSizeInBytes
        self.all_strands = []
        self.blockIndexSize = blockIndexSize
        self.intraBlockIndexSize = intraBlockIndexSize
        self._Policy = Policy
        return


    def _layered_decode_phys_to_strand(self, phys_strand):
        # perform block encoding
        try:
            phys_s = self.physCodec.decode(phys_strand)
            cw_s = self.physToStrandCodec.decode(phys_s)
        except err.DNAStorageError as p:
            cw_s = [-1] + [ 0 for _ in range(self.strandSizeInBytes-1) ]
            return cw_s
        
        try:
            s = self.strandCodec.decode(cw_s)
        except err.DNAStorageError as p:
            s = [-1] + [ 0 for _ in range(self.strandSizeInBytes-1) ]
        
        return s

    def decode_from_phys_to_strand(self, s):
        return self._layered_decode_phys_to_strand(s)
    
    def decode(self, phys_strand, bypass=False, input_key=None, input_value=None):
        if bypass==False:
            try:            
                s = self._layered_decode_phys_to_strand(phys_strand)
                self.all_strands.append(s)
            except err.DNAMissingPrimer as p:
                # just ignore strands that don't have a required primer
                pass
            except err.DNAStorageError as e:
                if self._Policy.allow(e):
                    pass
                else:
                    raise e
        else:
            self.all_strands.append(input_value)

    def _attempt_final_decoding(self):
        # do voting here!!        
        self.voted_strands = doMajorityVote(self.all_strands,\
                                            self.blockIndexSize+self.intraBlockIndexSize)
        if self.blockIndexSize>0:
            blocks = partitionStrandsIntoBlocks(self.all_strands,self.blockIndexSize)
        else:
            blocks = [ (0,self.all_strands) ]
        #print "number of blocks=",len(blocks)

        reportBlockStatus(blocks,self.minIndex,
                          self.blockIndexSize,self.intraBlockIndexSize)
        
        #blocks.sort()
        for b in blocks:
            idx = b[0]
            if idx < self.minIndex or idx >= self._packetizedFile.maxKey:
                # this happens due to errors in strands, and we should just
                # discard these erroneous blocks
                # If we want to reclaim some of these strands, it needs
                # to happen as part of the error processing per strand
                stats.inc("LayeredCodec::_attempt_final_decoding::indexOutOfRange")
                continue

            stats.inc("LayeredCodec::_attempt_final_decoding::numberOfBlocks")
            try:
                #print "attempt",idx,len(b[1])
                b_contig = self.strandToBlockCodec.decode(b)
                #print "attempt",b_contig[0],len(b_contig[1])
                b_noecc = self.blockCodec.decode(b_contig)
                #print "attempt",b_noecc[0],len(b_noecc[1])
                self.writeToFile(b_noecc[0],b_noecc[1])

            except err.DNAStorageError as e:
                if self._Policy.allow(e):
                    continue                
                else:
                    raise e
            except Exception as e:
                print (str(e))
                print (b)
                
            
    def write(self):
        self._attempt_final_decoding()
        super(LayeredDecoder,self).write()
            
    def dummy_write(self):
        self._attempt_final_decoding()
        return self._packetizedFile.dummy_write()
        
if __name__ == "__main__":
    from dnastorage.codec.rscodec import ReedSolomonOuterCodec
    from dnastorage.codec.strand import ReedSolomonInnerCodec
    from dnastorage.exceptions import NoTolerance
    from dnastorage.codec.commafreecodec import *
    from dnastorage.codec.phys import *
    
    pol = NoTolerance()

    payload=9
    
    outerECC = 70 # strands
    blockSize = payload*185 # 185 normal strands
    blockCodec = ReedSolomonOuterCodec(packetSize=blockSize,errorSymbols=70,payloadSize=payload,Policy=pol)

    blockToStrand = BlockToStrand(9,(185+70)*payload,Policy=pol)
    
    # take index into account
    # better to just say number of error symbols
    strandCodec = ReedSolomonInnerCodec(2,Policy=pol)

    codewords = CommaFreeCodewords(14)

    cut = InsertMidSequence('AGGTACCA')
    pre = PrependSequence('CAGGTACGCAGTTAGCACTC',CodecObj=cut, isPrimer=True)
    app = AppendSequence( 'CGTGGCAATATGACTACGGA',CodecObj=pre, isPrimer=True)

    flank = PrependSequence('CAGGAGAATGCCTTCCTAGG',CodecObj=app, isPrimer=True)
    flank = AppendSequence('CCTCGGTTCTTCTTGACCAG',CodecObj=flank, isPrimer=True)
    
    physCodec = flank
    pf = ReadPacketizedFile(sys.argv[1])
    
    enc = LayeredEncoder(pf,blockSizeInBytes=blockSize,strandSizeInBytes=payload,\
                         blockCodec=blockCodec,\
                         blockToStrandCodec=blockToStrand,\
                         strandCodec=strandCodec,\
                         strandToCodewordCodec=codewords,\
                         codewordToPhysCodec=CombineCodewords(),\
                         physCodec=physCodec)

    s = enc.dummy_encode()
    print (s[10])
    print (len(s),len(s[0]))

    all_strands = [ _ for _ in enc ]
    
    wpf = WritePacketizedFile(sys.argv[2],pf.size,blockSize)

    dec = LayeredDecoder(wpf,blockSizeInBytes=blockSize,strandSizeInBytes=payload,
                         blockCodec=blockCodec,\
                         strandCodec=strandCodec,\
                         physCodec=physCodec,\
                         physToStrandCodec=codewords,\
                         strandToBlockCodec=blockToStrand,\
                         Policy=pol)

    
    for ss in all_strands:
        for s in ss:
            dec.decode(s)

    dec._attempt_final_decoding()            
    dec.write()
    assert dec.complete

    



from dnastorage.exceptions import NoTolerance, AllowAll
from dnastorage.codec import commafreecodec
from dnastorage.codec.builder import customize_RS_CFC8

def ENC_FSMD_200(pf, primer5, primer3, bIndex=0, policy=NoTolerance(),withCut=None):
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,0,2,15,90,policy,withCut=withCut,outerECCStrands=20,minIndex=bIndex)
    return enc    

def DEC_FSMD_200(pf, primer5, primer3, bIndex=0,policy=AllowAll(),withCut=None):
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,0,2,15,90,policy,withCut=withCut,outerECCStrands=20,minIndex=bIndex)
    return dec

def ENC_RS_CFC8_200(pf, primer5, primer3, bIndex=0, policy=NoTolerance()):
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,2,2,15,185*15,policy,minIndex=bIndex)
    return enc    

def DEC_RS_CFC8_200(pf, primer5, primer3, bIndex=0, policy=AllowAll()):
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,2,2,15,185*15,policy,minIndex=bIndex)
    return dec    

FileSystemFormats = {
    # KEY      KEY    LEN  PacketSize, Abbrev.   Description                   ENCODER       DECODER
    0x0010 : [0x0010, 200, 90, "FSMD", "File system meta-data format", ENC_FSMD_200, DEC_FSMD_200 ],
    0x0020 : [0x0020, 200, 16, "RS+CFC8", "Reed-Solomon coded with ad hoc Comma-free codewords",
              ENC_RS_CFC8_200, DEC_RS_CFC8_200 ],
}

def file_system_formats():
    return [ v[3] for k,v in FileSystemFormats.items() ]

_abbrevFileSystemDict = { v[3] : v for k,v in FileSystemFormats.items() }

def file_system_format_description(formatid):
    return FileSystemFormats[formatid][4]

def file_system_format_packetsize(formatid):
    return FileSystemFormats[formatid][2]

def file_system_encoder(formatid):
    return FileSystemFormats[formatid][5]

def file_system_decoder(formatid):
    return FileSystemFormats[formatid][6]

def file_system_encoder_by_abbrev(ab):
    return _abbrevFileSystemDict[ab][5]

def file_system_decoder_by_abbrev(ab):
    return _abbrevFileSystemDict[ab][6]

def file_system_formatid_by_abbrev(ab):
    return _abbrevFileSystemDict[ab][0]



